# Implementation of the ideas presented in
# "Trace-Based Reconstruction of Affine Codes",
# by Gabriel Rodriguez, Jose M. Andion, Mahmut T. Kandemir, and Juan Tourino,
# to be presented in CGO 2016.
#
# This file is provided for Artifact Evaluation purposes only.
#
# Copyright: Gabriel Rodriguez (grodriguez@udc.es)

import numpy as np
import sys, time

def read_trace( path ):
	"""Read a memory trace from path. Since memory traces are usually large and
	bzip2 decompression is slow, we recommend to use generate_trace() instead.
	This should be used only if the reviewer wants to check that synthetically
	generated traces are equivalent to the real ones."""

	pf = open( path )
	a = []
	for l in pf:
		a += [ int(l.split()[1],16) ]
	
	a = np.array( a, dtype=np.int64 )
	c = [ a[1] - a[0] ]
	d = a[2:] - a[1:-1]
	pf.close()

	return (a,c,d)

def print_loop( base_access, c, U, w ):
	"""Internal function, used for result checking."""

        l = ""
        for x in xrange(len(U)):
                for y in range(x): l += "\t"
                l += "for x" + str(x)  + " in range("
                for y in xrange(x):
                        if U[x,y] != 0: l += str(int(U[x,y])) + "*x" + str(y) + " + "
                l += str(int(w[x])+1) + "):"
                l += "\n"
        tabs = ""
        for x in range(len(c)): tabs += "\t"
        l += tabs + "ar[pos] = a[" + str(int(base_access)) + "] +"
        for x in xrange(len(c)):
                l += str(int(c[x])) + "*x" + str(x) + " + "
        l = l[:-2] + "\n"
        l += tabs + "pos += 1\n"

        return l

def generate_trace( test="3mm", size="standard", access=0 ):
	"""This function generates a synthetic trace equivalent to a real PolyBench/C
	v3.2 one. Just set test=<application name> and access=<number>. Avaliable
	application names are self-documented in the function code, and are the same
	as application source names in PolyBench. We also provide an annotated
	PolyBench C/3.2 source so that the reviewer can check how, for instance,
	trmm.2 corresponds to a real access in the application code.

	By default, this function generates a trace of "standard" size (standard
	having the same meaning as compiling with -DSTANDARD_DATASET in PolyBench).
	In order to support quick functional validation, it is possible to use
	size="small" (equivalent to -DSMALL_DATASET) and size="mini" (equivalent
	to -DMINI_DATASET).

	The function returns three arrays: a (which contains the trace), d (which
	contains the first order differences of the trace), and c (which contains
	the stride of the second access with respect to the first, and constitutes
	the first known coefficient of loop index as described in Sec. 3 in the
	paper)."""

	a=[]

#       trmm
	if test == "trmm":
		ni = {
			"mini": 32,
			"small": 128,
			"standard": 1024,
		}[size]
		f = {
			0: lambda i,j,k: i*ni+j,
			1: lambda i,j,k: i*ni+k,
			2: lambda i,j,k: j*ni+k,
			3: lambda i,j,k: i*ni+j,
		}[access]
		for i in xrange(1,ni):
			for j in xrange(ni):
				for k in xrange(i):
					a += [ f(i,j,k) ]
#	trisolv
	elif test == "trisolv":
		n = {
			"mini": 32,
			"small": 500,
			"standard": 4000,
		}[size]
		f = {
			0: lambda i,j: i,
			1: lambda i,j: i,
			2: lambda i,j: i,
			3: lambda i,j: i*n+j,
			4: lambda i,j: j,
			5: lambda i,j: i,
			6: lambda i,j: i,
			7: lambda i,j: i*n+i,
			8: lambda i,j: i,
		}[access]
		if access in [0,1,6,7,8]:
			   for i in xrange(n):
				   a += [f(i,0)]
		else:
			for i in xrange(n):
				for j in xrange(i):
					a += [f(i,j)]
#	syrk
	elif test == "syrk":
		(ni,nj) = {
			"mini": (32,32),
			"small": (128,128),
			"standard": (1024,1024),
		}[size]
		f = {
			0: lambda i,j,k: i*ni+j,
			1: lambda i,j,k: i*ni+j,
			2: lambda i,j,k: i*nj+k,
			3: lambda i,j,k: j*nj+k,
			4: lambda i,j,k: i*ni+j,
			5: lambda i,j,k: i*ni+j,
		}[access]
		if access in [0,1]:
			for i in xrange(ni):
				for j in xrange(ni):
					a += [ f(i,j,0) ]
		else:
			for i in xrange(ni):
				for j in xrange(ni):
					for k in xrange(nj):
						a += [ f(i,j,k) ]
#	syr2k
	elif test == "syr2k":
		(ni,nj) = {
			"mini": (32,32),
			"small": (128,128),
			"standard": (1024,1024),
		}[size]
		f = {
			0: lambda i,j,k: i*ni+j,
			1: lambda i,j,k: i*ni+j,
			2: lambda i,j,k: i*ni+j,
			3: lambda i,j,k: i*nj+k,
			4: lambda i,j,k: j*nj+k,
			5: lambda i,j,k: i*ni+j,
			6: lambda i,j,k: i*ni+j,
			7: lambda i,j,k: i*nj+k,
			8: lambda i,j,k: j*nj+k,
			9: lambda i,j,k: i*ni+j,
		}[access]
		if access in [0,1]:
			for i in xrange(ni):
				for j in xrange(ni):
					a += [f(i,j,0)]
		else:
			for i in xrange(ni):
				for j in xrange(ni):
					for k in xrange(nj):
						a += [ f(i,j,k) ]

#	symm
	elif test == "symm":
		(ni,nj) = {
			"mini": (32,32),
			"small": (128,128),
			"standard": (1024,1024),
		}[size]
		f = {
			0: lambda i,j,k: k*nj+j,
			1: lambda i,j,k: k*nj+i,
			2: lambda i,j,k: i*nj+j,
			3: lambda i,j,k: k*nj+j,
			4: lambda i,j,k: k*nj+j,
			5: lambda i,j,k: k*nj+i,
			6: lambda i,j,k: i*nj+j,
			7: lambda i,j,k: i*nj+i,
			8: lambda i,j,k: i*nj+j,
			9: lambda i,j,k: i*nj+j,
		}[access]
		if access in [6,7,8,9]:
			   for i in xrange(ni):
				   for j in xrange(nj):
					   a += [ f(i,j,0) ]
		else:
			for i in xrange(ni):
				for j in xrange(nj):
					for k in xrange(j-1):
						a += [ f(i,j,k) ]
#	mvt
	elif test == "mvt":
		n = {
			"mini": 32,
			"small": 500,
			"standard": 4000,
		}[size]
		f = {
			0: lambda i,j: i,
			1: lambda i,j: i*n+j,
			2: lambda i,j: j,
			3: lambda i,j: i,
			4: lambda i,j: i,
			5: lambda i,j: j*n+i,
			6: lambda i,j: j,
			7: lambda i,j: i,
		}[access]
		for i in xrange(n):
			for j in xrange(n):
				a += [f(i,j)]
#	gesummv
	elif test == "gesummv":
		n = {
			"mini": 32,
			"small": 500,
			"standard": 4000,
		}[size]
		f = {
			0: lambda i,j: i,
			1: lambda i,j: i,
			2: lambda i,j: i*n+j,
			3: lambda i,j: j,
			4: lambda i,j: i,
			5: lambda i,j: i,
			6: lambda i,j: i*n+j,
			7: lambda i,j: j,
			8: lambda i,j: i,
			9: lambda i,j: i,
			10: lambda i,j: i,
			11: lambda i,j: i,
			12: lambda i,j: i,
		}[access]
		if access in [0,1,10,11,12]:
			for i in xrange(n):
				a += [f(i,0)]
		else:
			for i in xrange(n):
				for j in xrange(n):
					a += [f(i,j)]
#	gemver
	elif test == "gemver":
		n = {
			"mini": 32,
			"small": 500,
			"standard": 4000,
		}[size]
		f = {
			0: lambda i,j: i*n+j,
			1: lambda i,j: i,
			2: lambda i,j: j,
			3: lambda i,j: i,
			4: lambda i,j: j,
			5: lambda i,j: i*n+j,
			6: lambda i,j: i,
			7: lambda i,j: j*n+i,
			8: lambda i,j: j,
			9: lambda i,j: i,
			10: lambda i,j: i,
			11: lambda i,j: i,
			12: lambda i,j: i,
			13: lambda i,j: i,
			14: lambda i,j: i*n+j,
			15: lambda i,j: j,
			16: lambda i,j: i,
		}[access]
		if access in [10,11,12]:
			for i in xrange(n):
				a += [f(i,0)]
		else:
			for i in xrange(n):
				for j in xrange(n):
					a += [f(i,j)]
#	gemm
	elif test == "gemm":
		(ni,nj,nk) = {
			"mini": (32,32,32),
			"small": (128,128,128),
			"standard": (1024,1024,1024),
		}[size]
		f = {
			0: lambda i,j,k: i*nj+j,
			1: lambda i,j,k: i*nj+j,
			2: lambda i,j,k: i*nj+j,
			3: lambda i,j,k: i*nk+k,
			4: lambda i,j,k: k*nj+j,
			5: lambda i,j,k: i*nj+j,
		}[access]
		if access in [0,1]:
			for i in xrange(ni):
				for j in xrange(nj):
					a += [f(i,j,0)]
		else:
			for i in xrange(ni):
				for j in xrange(nj):
					for k in xrange(nk):
						a += [f(i,j,k)]
#	doitgen
	elif test == "doitgen":
		(nq,nr,npp) = {
			"mini": (10,10,10),
			"small": (32,32,32),
			"standard": (128,128,128),
		}[size]
		f = {
			0: lambda r,q,p,s: r*nq*npp+q*npp+p,
			1: lambda r,q,p,s: r*nq*npp+q*npp+p,
			2: lambda r,q,p,s: r*nq*npp+q*npp+s,
			3: lambda r,q,p,s: s*npp+p,
			4: lambda r,q,p,s: r*nq*npp+q*npp+p,
			5: lambda r,q,p,s: r*nq*npp+q*npp+p,
			6: lambda r,q,p,s: r*nq*npp+q*npp+p,
		}[access]
		if access in [0,5,6]:
			for r in xrange(nr):
				for q in xrange(nq):
					for p in xrange(npp):
						a += [f(r,q,p,0)]
		else:
			for r in xrange(nr):
				for q in xrange(nq):
					for p in xrange(npp):
						for s in xrange(npp):
							a += [f(r,q,p,s)]
#	cholesky
	elif test == "cholesky":
		n = {
			"mini": 32,
			"small": 128,
			"standard": 1024,
		}[size]
		f = {
			0: lambda i,j,k: i*n+i,
			1: lambda i,j,k: i*n+j,
			2: lambda i,j,k: i,
			3: lambda i,j,k: i*n+j,
			4: lambda i,j,k: j*n+k,
			5: lambda i,j,k: i*n+k,
			6: lambda i,j,k: i,
			7: lambda i,j,k: j*n+i,
		}[access]
		if access in [0,2]:
			for i in xrange(n):
				a += [f(i,0,0)]
		elif access in [1]:
			for i in xrange(n):
				for j in xrange(i):
					a += [f(i,j,0)]
		elif access in [3,6,7]:
			for i in xrange(n):
				for j in xrange(i+1,n):
					a += [f(i,j,0)]
		else:
			for i in xrange(n):
				for j in xrange(i+1,n):
					for k in xrange(0,i):
						a += [f(i,j,k)]
#	bicg
	elif test == "bicg":
		(nx,ny) = {
			"mini": (32,32),
			"small": (500,500),
			"standard": (4000,4000),
		}[size]
		f = {
			0: lambda i,j: i,
			1: lambda i,j: i,
			2: lambda i,j: j,
			3: lambda i,j: i,
			4: lambda i,j: i*ny+j,
			5: lambda i,j: j,
			6: lambda i,j: i,
			7: lambda i,j: i*ny+j,
			8: lambda i,j: j,
			9: lambda i,j: i,
		}[access]
		if access in [0,1]:
			for i in xrange(nx):
				a += [f(i,0)]
		else:
			for i in xrange(nx):
				for j in xrange(ny):
					a += [f(i,j)]
#	atax
	elif test == "atax":
		(nx,ny) = {
			"mini": (32,32),
			"small": (500,500),
			"standard": (4000,4000),
		}[size]
		f = {
			0: lambda i,j: i,
			1: lambda i,j: i,
			2: lambda i,j: i,
			3: lambda i,j: i*ny+j,
			4: lambda i,j: j,
			5: lambda i,j: i,
			6: lambda i,j: j,
			7: lambda i,j: i*ny+j,
			8: lambda i,j: i,
			9: lambda i,j: j,
		}[access]
		if access in [0,1]:
			for i in xrange(nx):
				a += [f(i,0)]
		else:
			for i in xrange(nx):
				for j in xrange(ny):
					a += [f(i,j)]
#	3mm
	elif test == "3mm":
		(ni,nj,nk,nl,nm) = {
			"mini": (32,32,32,32,32),
			"small": (128,128,128,128,128),
			"standard": (1024,1024,1024,1024,1024),
		}[size]
		f = {
			0: lambda i,j,k: i*nj+j,
			1: lambda i,j,k: i*nj+j,
			2: lambda i,j,k: i*nk+k,
			3: lambda i,j,k: k*nj+j,
			4: lambda i,j,k: i*nj+j,
			5: lambda i,j,k: i*nl+j,
			6: lambda i,j,k: i*nl+j,
			7: lambda i,j,k: i*nm+k,
			8: lambda i,j,k: k*nl+j,
			9: lambda i,j,k: i*nl+j,
			10: lambda i,j,k: i*nl+j,
			11: lambda i,j,k: i*nl+j,
			12: lambda i,j,k: i*nj+k,
			13: lambda i,j,k: k*nl+j,
			14: lambda i,j,k: i*nl+j,
		}[access]
		if access in [0,5,10]:
			for i in xrange(ni):
				for j in xrange(nj):
					a += [f(i,j,0)]
		else:
			for i in xrange(ni):
				for j in xrange(nj):
					for k in xrange(nk):
						a += [f(i,j,k)]
#	2mm
	elif test == "2mm":
		(ni,nj,nk,nl) = {
			"mini": (32,32,32,32),
			"small": (128,128,128,128),
			"standard": (1024,1024,1024,1024),
		}[size]
		f = {
			0: lambda i,j,k: i*nj+j,
			1: lambda i,j,k: i*nj+j,
			2: lambda i,j,k: i*nk+k,
			3: lambda i,j,k: k*nj+j,
			4: lambda i,j,k: i*nj+j,
			5: lambda i,j,k: i*nl+j,
			6: lambda i,j,k: i*nl+j,
			7: lambda i,j,k: i*nl+j,
			8: lambda i,j,k: i*nj+k,
			9: lambda i,j,k: k*nj+j,
			10: lambda i,j,k: i*nl+j,
		}[access]
		if access in [0,5,6]:
			for i in xrange(ni):
				for j in xrange(nj):
					a += [f(i,j,0)]
		else:
			for i in xrange(ni):
				for j in xrange(nl):
					for k in xrange(nj):
						a += [f(i,j,k)]
#	ludcmp
	elif test == "ludcmp":
		n = {
			"mini": 32,
			"small": 128,
			"standard": 1024,
		}[size]
		f = {
			0: lambda i,j,k: j*n+i,
			1: lambda i,j,k: j*n+k,
			2: lambda i,j,k: k*n+i,
			3: lambda i,j,k: i*n+i,
			4: lambda i,j,k: j*n+i,
			5: lambda i,j,k: (i+1)*n+j,
			6: lambda i,j,k: (i+1)*n+k,
			7: lambda i,j,k: k*n+j,
			8: lambda i,j,k: (i+1)*n+j,
			9: lambda i,j,k: i,
			10: lambda i,j,k: i*n+j,
			11: lambda i,j,k: j,
			12: lambda i,j,k: i,
			13: lambda i,j,k: n-1-i,
			14: lambda i,j,k: (n-1-i)*n+j,
			15: lambda i,j,k: j,
			16: lambda i,j,k: (n-1-i)*n+n-1-i,
			17: lambda i,j,k: n-1-i,
		}[access]
		if access in [0,3,4,5,8]:
			for i in xrange(n):
				for j in xrange(i+1,n+1):
					a += [f(i,j,0)]
		elif access in [1,2]:
			for i in xrange(n):
				for j in xrange(i+1,n+1):
					for k in xrange(i):
						a += [f(i,j,k)]
		elif access in [9,12]:
			for i in xrange(1,n+1):
				a += [f(i,0,0)]
		elif access in [10,11]:
			for i in xrange(1,n+1):
				for j in xrange(i):
					a += [f(i,j,0)]
		elif access in [13,16,17]:
			for i in xrange(n):
				a += [f(i,0,0)]
		elif access in [14,15]:
			for i in xrange(n):
				for j in xrange(n-i,n+1):
					a += [f(i,j,0)]
		else:
			for i in xrange(n):
				for j in xrange(i+1,n+1):
					for k in xrange(i+1):
						a += [f(i,j,k)]
#	lu
	elif test == "lu":
		n = {
			"mini": 32,
			"small": 128,
			"standard": 1024,
		}[size]
		f = {
			0: lambda i,j,k: k*n+j,
			1: lambda i,j,k: k*n+k,
			2: lambda i,j,k: k*n+j,
			3: lambda i,j,k: i*n+j,
			4: lambda i,j,k: i*n+k,
			5: lambda i,j,k: k*n+j,
			6: lambda i,j,k: i*n+j,
		}[access]
		if access in [0,1,2]:
			for k in xrange(n):
				for j in xrange(k+1,n):
					a += [f(0,j,k)]
		else:
			for k in xrange(n):
				for i in xrange(k+1,n):
					for j in xrange(k+1,n):
						a += [f(i,j,k)]
# 	gramschmidt
	elif test == "gramschmidt":
		(ni,nj) = {
			"mini": (32,32),
			"small": (128,128),
			"standard": (512,512),
		}[size]
		f = {
			0: lambda i,j,k: i*nj+k,
			1: lambda i,j,k: k*nj+k,
			2: lambda i,j,k: i*nj+k,
			3: lambda i,j,k: k*nj+k,
			4: lambda i,j,k: i*nj+k,
			5: lambda i,j,k: k*nj+j,
			6: lambda i,j,k: k*nj+j,
			7: lambda i,j,k: i*nj+k,
			8: lambda i,j,k: i*nj+j,
			9: lambda i,j,k: k*nj+j,
			10: lambda i,j,k: i*nj+j,
			11: lambda i,j,k: i*nj+k,
			12: lambda i,j,k: k*nj+j,
			13: lambda i,j,k: i*nj+j,
		}[access]
		if access in [0,2,3,4]:
			   for k in xrange(nj):
				   for i in xrange(ni):
					   a += [f(i,0,k)]
		elif access in [1]:
			for k in xrange(nj):
				a += [f(0,0,k)]
		elif access in [5]:
			for k in xrange(nj):
				for j in xrange(k+1,nj):
					a += [f(0,j,k)]
		else: 
			for k in xrange(nj):
				for j in xrange(k+1,nj):
					for i in xrange(ni):
						a += [f(i,j,k)]
#	dynprog
	elif test == "dynprog":
		(tsteps,length) = {
			"mini": (10,32),
			"small": (100,50),
			"standard": (10000,50),
		}[size]
		f = {
			0: lambda i,j,k: i*length+j,
			1: lambda i,j,k: i*length*length+j*length+i,
			2: lambda i,j,k: i*length*length + j*length + k - 1,
			3: lambda i,j,k: i*length+k,
			4: lambda i,j,k: k*length+j,
			5: lambda i,j,k: i*length*length + j*length + k,
			6: lambda i,j,k: i*length*length + j*length + j - 1,
			7: lambda i,j,k: i*length + j,
			8: lambda i,j,k: i*length + j,
		}[access]
		if access in [0]:
			for t in xrange(tsteps):
				for i in xrange(length):
					for j in xrange(length):
						a += [f(i,j,0)]
		elif access in [1,6,7,8]:
			for t in xrange(tsteps):
				for i in xrange(length-1):
					for j in xrange(i+1,length):
						a += [f(i,j,0)]
		else:
			for t in xrange(tsteps):
				for i in xrange(length-1):
					for j in xrange(i+1,length):
						for k in xrange(i+1,j):
							a += [f(i,j,k)]
#	durbin
	elif test == "durbin":
		n = {
			"mini": 32,
			"small": 500,
			"standard": 4000,
		}[size]
		f = {
			0: lambda i,k: k-1,
			1: lambda i,k: k-1,
			2: lambda i,k: k,
			3: lambda i,k: k,
			4: lambda i,k: k,
			5: lambda i,k: i*n+k,
			6: lambda i,k: k-i-1,
			7: lambda i,k: i*n+k-1,
			8: lambda i,k: (i+1)*n+k,
			9: lambda i,k: k*n+k,
			10: lambda i,k: k,
			11: lambda i,k: k,
			12: lambda i,k: i*n+k-1,
			13: lambda i,k: k,
			14: lambda i,k: (k-i-1)*n+k-1,
			15: lambda i,k: i*n+k,
			16: lambda i,k: k,
			17: lambda i,k: k*n+k,
			18: lambda i,k: i*n+n,
			19: lambda i,k: i,
		}[access]
		if access in [0,1,2,3,4,9,10,11,16,17]:
			for k in xrange(1,n):
				a += [f(0,k)]
		elif access in [18,19]:
			for i in xrange(n):
				a += [f(i,0)]
		else:
			for k in xrange(1,n):
				for i in xrange(0,k):
					a += [f(i,k)]
#	adi
	elif test == "adi":
		(tsteps,n) = {
			"mini": (2,32),
			"small": (10,500),
			"standard": (50,1024),
		}[size]
		f = {
			0: lambda t,i1,i2: i1*n+i2,
			1: lambda t,i1,i2: i1*n+i2-1,
			2: lambda t,i1,i2: i1*n+i2,
			3: lambda t,i1,i2: i1*n+i2-1,
			4: lambda t,i1,i2: i1*n+i2,
			5: lambda t,i1,i2: i1*n+i2,
			6: lambda t,i1,i2: i1*n+i2,
			7: lambda t,i1,i2: i1*n+i2-1,
			8: lambda t,i1,i2: i1*n+i2,
			9: lambda t,i1,i2: i1*n+n-1,
			10:lambda t,i1,i2: i1*n+n-1,
			11:lambda t,i1,i2: i1*n+n-1,
			12: lambda t,i1,i2: i1*n+n-2-i2,
			13: lambda t,i1,i2: i1*n+n-2-i2-1,
			14: lambda t,i1,i2: i1*n+n-i2-3,
			15: lambda t,i1,i2: i1*n+n-3-i2,
			16: lambda t,i1,i2: i1*n+n-i2-2,
			17: lambda t,i1,i2: i1*n+i2,
			18: lambda t,i1,i2: (i1-1)*n+i2,
			19: lambda t,i1,i2: i1*n+i2,
			20: lambda t,i1,i2: (i1-1)*n+i2,
			21: lambda t,i1,i2: i1*n+i2,
			22: lambda t,i1,i2: i1*n+i2,
			23: lambda t,i1,i2: i1*n+i2,
			24: lambda t,i1,i2: (i1-1)*n+i2,
			25: lambda t,i1,i2: i1*n+i2,
			26: lambda t,i1,i2: (n-1)*n+i2,
			27: lambda t,i1,i2: (n-1)*n+i2,
			28: lambda t,i1,i2: (n-1)*n+i2,
			29: lambda t,i1,i2: (n-2-i1)*n+i2,
			30: lambda t,i1,i2: (n-i1-3)*n+i2,
			31: lambda t,i1,i2: (n-3-i1)*n+i2,
			32: lambda t,i1,i2: (n-2-i1)*n+i2,
			33: lambda t,i1,i2: (n-2-i1)*n+i2,
		}[access]
		if access in [9,10,11]:
			for t in xrange(tsteps):
				for i1 in xrange(n):
					a += [f(t,i1,0)]
		elif access in [12,13,14,15,16]:
			for t in xrange(tsteps):
				for i1 in xrange(n):
					for i2 in xrange(n-2):
						a += [f(t,i1,i2)]
		elif access in [26,27,28]:
			for t in xrange(tsteps):
				for i2 in xrange(n):
					a += [f(t,0,i2)]
		elif access in [29,30,31,32,33]:
			for t in xrange(tsteps):
				for i1 in xrange(n-2):
					for i2 in xrange(n):
						a += [f(t,i1,i2)]
		else:
			for t in xrange(tsteps):
				for i1 in xrange(n):
					for i2 in xrange(n):
						a += [f(t,i1,i2)]
#	fdtd-2d
	elif test == "fdtd-2d":
		(tmax,nx,ny) = {
			"mini": (2,32,32),
			"small": (10,500,500),
			"standard": (50,1000,1000),
		}[size]
		f = {
			0: lambda t,i,j: t,
			1: lambda t,i,j: j,
			2: lambda t,i,j: i*ny+j,
			3: lambda t,i,j: i*ny+j,
			4: lambda t,i,j: (i-1)*ny+j,
			5: lambda t,i,j: i*ny+j,
			6: lambda t,i,j: i*ny+j,
			7: lambda t,i,j: i*ny+j,
			8: lambda t,i,j: i*ny+j-1,
			9: lambda t,i,j: i*ny+j,
			10: lambda t,i,j: i*ny+j,
			11: lambda t,i,j: i*ny+j+1,
			12: lambda t,i,j: i*ny+j,
			13: lambda t,i,j: (i+1)*ny+j,
			14: lambda t,i,j: i*ny+j,
			15: lambda t,i,j: i*ny+j,
		}[access]
		if access in [0,1]:
			for t in xrange(tmax):
				for j in xrange(ny):
					a += [f(t,0,j)]
		elif access in [2,3,4,5]:
			for t in xrange(tmax):
				for i in xrange(1,nx):
					for j in xrange(ny):
						a += [f(t,i,j)]
		elif access in [6,7,8,9]:
			for t in xrange(tmax):
				for i in xrange(nx):
					for j in xrange(1,ny):
						a += [f(t,i,j)]
		else:
			for t in xrange(tmax):
				for i in xrange(nx):
					for j in xrange(ny):
						a += [f(t,i,j)]
#	fdtd-apml
	elif test == "fdtd-apml":
		(cz,cym,cxm) = {
			"mini": (32,32,32),
			"small": (64,64,64),
			"standard": (256,256,256),
		}[size]
		f = {
			0: lambda iz,iy,ix: iz*cz+iy,
			1: lambda iz,iy,ix: iz*cz**2+iy*cz+ix,
			2: lambda iz,iy,ix: iz*cz**2+(iy+1)*cz+ix,
			3: lambda iz,iy,ix: iz*cz**2+iy*cz+ix+1,
			4: lambda iz,iy,ix: iz*cz**2+iy*cz+ix,
			5: lambda iz,iy,ix: iz*cz+iy,
			6: lambda iz,iy,ix: iy,
			7: lambda iz,iy,ix: iy,
			8: lambda iz,iy,ix: iz*cz**2+iy*cz+ix,
			9: lambda iz,iy,ix: iy,
			10: lambda iz,iy,ix: iz*cz+iy,
			11: lambda iz,iy,ix: iz*cz**2+iy*cz+ix,
			12: lambda iz,iy,ix: ix,
			13: lambda iz,iy,ix: ix,
			14: lambda iz,iy,ix: iz*cz**2+iy*cz+ix,
			15: lambda iz,iy,ix: iz,
			16: lambda iz,iy,ix: ix,
			17: lambda iz,iy,ix: iz*cz+iy,
			18: lambda iz,iy,ix: iz,
			19: lambda iz,iy,ix: ix,
			20: lambda iz,iy,ix: iz*cz**2+iy*cz+ix,
			21: lambda iz,iy,ix: iz*cz**2+iy*cz+ix,
			22: lambda iz,iy,ix: iz*cz+iy,
			23: lambda iz,iy,ix: iz*cz+iy,
			24: lambda iz,iy,ix: iz*cz**2+iy*cz+cz,
			25: lambda iz,iy,ix: iz*cz**2+(iy+1)*cz+cz,
			26: lambda iz,iy,ix: iz*cz+iy,
			27: lambda iz,iy,ix: iz*cz**2+iy*cz+cz,
			28: lambda iz,iy,ix: iz*cz+iy,
			29: lambda iz,iy,ix: iy,
			30: lambda iz,iy,ix: iy,
			31: lambda iz,iy,ix: iz*cz**2+iy*cz+cz,
			32: lambda iz,iy,ix: iy,
			33: lambda iz,iy,ix: iz*cz+iy,
			34: lambda iz,iy,ix: iz*cz**2+iy*cz+cz,
			35: lambda iz,iy,ix: cz,
			36: lambda iz,iy,ix: cz,
			37: lambda iz,iy,ix: iz*cz**2+iy*cz+cz,
			38: lambda iz,iy,ix: iz,
			39: lambda iz,iy,ix: cz,
			40: lambda iz,iy,ix: iz*cz+iy,
			41: lambda iz,iy,ix: iz,
			42: lambda iz,iy,ix: cz,
			43: lambda iz,iy,ix: iz*cz**2+iy*cz+cz,
			44: lambda iz,iy,ix: iz*cz**2+iy*cz+cz,
			45: lambda iz,iy,ix: iz*cz+iy,
			46: lambda iz,iy,ix: iz*cz+iy,
			47: lambda iz,iy,ix: iz*cz**2+cz*cz+ix,
			48: lambda iz,iy,ix: iz*cz+ix,
			49: lambda iz,iy,ix: iz*cz**2+cz*cz+ix+1,
			50: lambda iz,iy,ix: iz*cz**2+cz*cz+ix,
			51: lambda iz,iy,ix: iz*cz+iy,
			52: lambda iz,iy,ix: cz,
			53: lambda iz,iy,ix: iy,
			54: lambda iz,iy,ix: iz*cz**2+iy*cz+ix,
			55: lambda iz,iy,ix: iy,
			56: lambda iz,iy,ix: iz*cz+iy,
			57: lambda iz,iy,ix: iz*cz**2+cz*cz+ix,
			58: lambda iz,iy,ix: ix,
			59: lambda iz,iy,ix: ix,
			60: lambda iz,iy,ix: iz*cz**2+cz*cz+ix,
			61: lambda iz,iy,ix: iz,
			62: lambda iz,iy,ix: ix,
			63: lambda iz,iy,ix: iz*cz+iy,
			64: lambda iz,iy,ix: iz,
			65: lambda iz,iy,ix: ix,
			66: lambda iz,iy,ix: iz*cz**2+cz*cz+ix,
			67: lambda iz,iy,ix: iz*cz**2+cz*cz+ix,
			68: lambda iz,iy,ix: iz*cz+iy,
			69: lambda iz,iy,ix: iz*cz+iy,
			70: lambda iz,iy,ix: iz*cz**2+cz*cz+cz,
			71: lambda iz,iy,ix: iz*cz+cz,
			72: lambda iz,iy,ix: iz*cz+cz,
			73: lambda iz,iy,ix: iz*cz**2+cz*cz+cz,
			74: lambda iz,iy,ix: iz*cz+iy,
			75: lambda iz,iy,ix: cz,
			76: lambda iz,iy,ix: cz,
			77: lambda iz,iy,ix: iz*cz**2+cz*cz+cz,
			78: lambda iz,iy,ix: cz,
			79: lambda iz,iy,ix: iz*cz+iy,
			80: lambda iz,iy,ix: iz*cz**2+cz*cz+cz,
			81: lambda iz,iy,ix: cz,
			82: lambda iz,iy,ix: cz,
			83: lambda iz,iy,ix: iz*cz**2+cz*cz+cz,
			84: lambda iz,iy,ix: iz,
			85: lambda iz,iy,ix: cz,
			86: lambda iz,iy,ix: iz*cz+iy,
			87: lambda iz,iy,ix: iz,
			88: lambda iz,iy,ix: cz,
			89: lambda iz,iy,ix: iz*cz**2+cz*cz+cz,
			90: lambda iz,iy,ix: iz*cz**2+cz*cz+cz,
			91: lambda iz,iy,ix: iz*cz+iy,
		}[access]
		if access in [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68]:
			for iz in xrange(cz):
				for iy in xrange(cz):
					for ix in xrange(cz):
						a += [f(iz,iy,ix)]

		else:
			for iz in xrange(cz):
				for iy in xrange(cz):
					a += [f(iz,iy,0)]
#	jacobi-1d
	elif test == "jacobi-1d":
		(tsteps,n) = {
			"mini": (2,500),
			"small": (10,1000),
			"standard": (100,10000),
		}[size]
		f = {
			0: lambda t,i: i-1,
			1: lambda t,i: i,
			2: lambda t,i: i+1,
			3: lambda t,i: i,
			4: lambda t,i: i,
			5: lambda t,i: i,
		}[access]
		for t in xrange(tsteps):
			for i in xrange(n):
				a += [f(t,i)]
#	jacobi-2d
	elif test == "jacobi-2d":
		(tsteps,n) = {
			"mini": (2,32),
			"small": (10,500),
			"standard": (20,1000),
		}[size]
		f = {
			0: lambda i,j: i*n+j,
			1: lambda i,j: i*n+j-1,
			2: lambda i,j: i*n+1+j,
			3: lambda i,j: (1+i)*n+j,
			4: lambda i,j: (i-1)*n+j,
			5: lambda i,j: i*n+j,
			6: lambda i,j: i*n+j,
			7: lambda i,j: i*n+j,
		}[access]
		for t in xrange(tsteps):
			for i in xrange(1,n-1):
				for j in xrange(1,n-1):
					a += [f(i,j)]
#	seidel-2d
	elif test == "seidel-2d":
		(tsteps,n) = {
			"mini": (2,32),
			"small": (10,500),
			"standard": (20,1000),
		}[size]
		f = {
			0: lambda t,i,j: (i-1)*n+j-1,
			1: lambda t,i,j: (i-1)*n+j,
			2: lambda t,i,j: (i-1)*n+j+1,
			3: lambda t,i,j: i*n+j-1,
			4: lambda t,i,j: i*n+j,
			5: lambda t,i,j: i*n+j+1,
			6: lambda t,i,j: (i+1)*n+j-1,
			7: lambda t,i,j: (i+1)*n+j,
			8: lambda t,i,j: (i+1)*n+j+1,
			9: lambda t,i,j: i*n+j,
		}[access]
		for t in xrange(tsteps):
			for i in xrange(1,n-1):
				for j in xrange(1,n-1):
					a += [f(t,i,j)]
#	correlation
	elif test == "correlation":
		(n,m) = {
			"mini": (32,32),
			"small": (500,500),
			"standard": (1000,1000),
		}[size]
		f = {
			0: lambda i,j,k: j,
			1: lambda i,j,k: j,
			2: lambda i,j,k: i*n+j,
			3: lambda i,j,k: j,
			4: lambda i,j,k: j,
			5: lambda i,j,k: j,
			6: lambda i,j,k: j,
			7: lambda i,j,k: j,
			8: lambda i,j,k: i*n+j,
			9: lambda i,j,k: j,
			10: lambda i,j,k: j,
			11: lambda i,j,k: j,
			12: lambda i,j,k: j,
			13: lambda i,j,k: j,
			14: lambda i,j,k: j,
			15: lambda i,j,k: j,
			16: lambda i,j,k: j,
			17: lambda i,j,k: i*n+j,
			18: lambda i,j,k: j,
			19: lambda i,j,k: i*n+j,
			20: lambda i,j,k: i*n+j,
			21: lambda i,j,k: j,
			22: lambda i,j,k: i*n+j,
			23: lambda i,j,k: i*m+i,
			24: lambda i,j,k: i*m+j,
			25: lambda i,j,k: i*m+j,
			26: lambda i,j,k: k*n+i,
			27: lambda i,j,k: k*n+j,
			28: lambda i,j,k: i*m+j,
			29: lambda i,j,k: i*m+j,
			30: lambda i,j,k: j*m+i,
		}[access]
		if access in [0,4,5,6,11,12,13,14,15,16]:
			for j in xrange(m):
				a += [f(0,j,0)]
		if access in [1,2,3,7,8,9,10]:
			for j in xrange(m):
				for i in xrange(n):
					a += [f(i,j,0)]
		if access in [17,18,19,20,21,22]:
			for i in xrange(n):
				for j in xrange(m):
					a += [f(i,j,0)]
		if access in [23]:
			for i in xrange(m-1):
				a += [f(i,0,0)]
		if access in [24,29,30]:
			for i in xrange(m-1):
				for j in xrange(i+1,m):
					a += [f(i,j,0)]
		if access in [25,26,27,28]:
			for i in xrange(m-1):
				for j in xrange(i+1,m):
					for k in xrange(n):
						a += [f(i,j,k)]
#	covariance
	elif test == "covariance":
		(n,m) = {
			"mini": (32,32),
			"small": (500,500),
			"standard": (1000,1000),
		}[size]
		f = {
			0: lambda j1,j2,i: j1,
			1: lambda j1,j2,i: j1,
			2: lambda j1,j2,i: j1,
			3: lambda j1,j2,i: i*n+j1,
			4: lambda j1,j2,i: j1,
			5: lambda j1,j2,i: j1,
			6: lambda j1,j2,i: i*n+j1,
			7: lambda j1,j2,i: i*n+j1,
			8: lambda j1,j2,i: j1,
			9: lambda j1,j2,i: j1*n+j2,
			10: lambda j1,j2,i: j1*n+j2,
			11: lambda j1,j2,i: j1*n+j2,
			12: lambda j1,j2,i: i*n+j1,
			13: lambda j1,j2,i: i*n+j2,
			14: lambda j1,j2,i: j2*n+j1,
			15: lambda j1,j2,i: j1*n+j2,
		}[access]
		if access in [0,4,5]:
			for j1 in xrange(n):
				a += [f(j1,0,0)]
		elif access in [1,2,3]:
			for j1 in xrange(n):
				for i in xrange(n):
					a += [f(j1,0,i)]
		elif access in [6,7,8]:
			for i in xrange(n):
				for j1 in xrange(n):
					a += [f(j1,0,i)]
		elif access in [9,14,15]:
			for j1 in xrange(n):
				for j2 in xrange(j1,n):
					a += [f(j1,j2,0)]
		else:
			for j1 in xrange(n):
				for j2 in xrange(j1,n):
					for i in xrange(n):
						a += [f(j1,j2,i)]
#	floyd-warshall
	elif test == "floyd-warshall":
		n = {
			"mini": 32,
			"small": 128,
			"standard": 1024,
		}[size]
		f = {
			0: lambda i,j,k: i*n+j,
			1: lambda i,j,k: i*n+k,
			2: lambda i,j,k: k*n+j,
			3: lambda i,j,k: i*n+j,
		}[access]
		for k in xrange(n):
			for i in xrange(n):
				for j in xrange(n):
					a += [f(i,j,k)]
#	reg_detect
	elif test == "reg_detect":
		(niter,length,maxgrid) = {
			"mini": (10,32,2),
			"small": (100,50,6),
			"standard": (10000,64,6),
		}[size]
		f = {
			0: lambda t,j,i,cnt: j*length*maxgrid+i*length+cnt,
			1: lambda t,j,i,cnt: maxgrid*j+i,
			2: lambda t,j,i,cnt: length*maxgrid*j+length*i,
			3: lambda t,j,i,cnt: length*maxgrid*j+length*i,
			4: lambda t,j,i,cnt: length*maxgrid*j+length*i+cnt,
			5: lambda t,j,i,cnt: length*maxgrid*j+length*i+cnt-1,
			6: lambda t,j,i,cnt: length*maxgrid*j+length*i+cnt,
			7: lambda t,j,i,cnt: j*maxgrid+i,
			8: lambda t,j,i,cnt: length*maxgrid*j+length*i+length-1,
			9: lambda t,j,i,cnt: i,
			10: lambda t,j,i,cnt: i,
			11: lambda t,j,i,cnt: j*maxgrid+i,
			12: lambda t,j,i,cnt: (j-1)*maxgrid+i-1,
			13: lambda t,j,i,cnt: j*maxgrid+i,
		}[access]
		if access in [0,1]:
			for t in xrange(niter):
				for j in xrange(maxgrid):
					for i in xrange(j,maxgrid):
						for cnt in xrange(length):
							a += [f(t,j,i,cnt)]
		elif access in [2,3,7,8]:
			for t in xrange(niter):
				for j in xrange(maxgrid):
					for i in xrange(j,maxgrid):
						a += [f(t,j,i,0)]
		elif access in [4,5,6]:
			for t in xrange(niter):
				for j in xrange(maxgrid):
					for i in xrange(j,maxgrid):
						for cnt in xrange(1,length):
							a += [f(t,j,i,cnt)]
		elif access in [9,10]:
			for t in xrange(niter):
				for i in xrange(maxgrid):
					a += [f(t,0,i,0)]
		else:
			for t in xrange(niter):
				for j in xrange(1,maxgrid):
					for i in xrange(j,maxgrid):
						a += [f(t,j,i,0)]
	elif test == "exhaustive":
		n = {
			"mini": 32,
			"small": 128,
			"standard": 1024,
		}[size]
		
		for i in xrange(n):
			for j in xrange(i+1):
				for k in xrange(n-i):
					a += [n*i + k]

	a=np.array(a,dtype=np.int64)
	c=np.array([a[1]-a[0]], dtype=np.int64)
	d=a[2:]-a[1:-1]
	return (a,c,d)


# This dictionary contains the expected results of the reconstruction of
# each trace. It is used in automatic system tests. Its meaning is
# straightforward, e.g., library['trmm'][2] contains the expected loop
# depth of the reconstruction af access trmm.2. Note that expected
# reconstruction depth may be lower than the original loop depth in the
# code, as explained in the paper.
library = dict()
library['trmm'] = { 0: 3, 3:3,
		     1: 3, 
		     2: 3 }
library['trisolv'] = { 0:1, 1:1, 6:1, 8:1,
		       2:2, 5:2,
		       3:2,
		       4:2,
		       7:1 }
library['syrk'] = { 0:1, 1:1,
		    2:3,
		    3:2,
		    4:2, 5:2 }
library['syr2k'] = { 0:1, 1:1,
		     2:2, 5:2, 6:2, 9:2,
		     3:3, 7:3,
		     4:2, 8:2 }
library['symm'] = { 0:3, 3:3, 4:3,
		    1:3,
		    2:3, 5:3,
		    6:1, 8:1, 9:1,
		    7:2 }
library['mvt'] = { 0:2, 3:2, 4:2, 7:2,
		   1:1,
		   2:2, 6:2,
		   5:2 }
library['gesummv'] = { 0:1, 1:1, 10:1, 11:1, 12:1,
		       2:1, 6:1,
		       3:2, 7:2,
		       4:2, 5:2, 8:2, 9:2 }
library['gemver'] = { 0:1, 5:1, 14:1,
		      1:2, 3:2, 6:2, 9:2, 13:2, 16:2,
		      2:2, 4:2, 8:2, 15:2,
		      7:2,
		      10:1, 11:1, 12:1 }
library['gemm'] = { 0:1, 1:1,
		    2:2, 5:2,
		    3:3,
		    4:3 }
library['doitgen'] = { 0:1, 5:1, 6:1,
		       1:2, 4:2,
		       2:3, 3:3 }
library['cholesky'] = { 0:1,
			1:2,
			2:1,
			3:2,
			4:3,
			5:3,
			6:2,
			7:2 }
library['bicg'] = { 0:1, 1:1,
		    2:2, 5:2, 8:2,
		    3:2, 6:2, 9:2,
		    4:1, 7:1 }
library['atax'] = { 0:1, 1:1,
		    2:2, 5:2, 8:2,
		    3:1, 7:1,
		    4:2, 6:2, 9:2 }
library['3mm'] = { 0:1, 
		   1:2, 4:2,
		   2:3,
		   3:3, 
		   5:1, 10:1,
		   6:2, 9:2, 11:2, 14:2,
		   7:3, 
		   8:3, 13:3,
		   12:3 }
library['2mm'] = { 0:1,
		   1:2, 4:2,
		   2:3,
		   3:3, 9:3,
		   5:1, 6:1,
		   7:2, 10:2,
		   8:3 }
library['ludcmp'] = { 0:2, 4:2,
		      1:3,
		      2:3,
		      3:2,
		      5:2, 8:2,
		      6:3,
		      7:3,
		      9:1, 12:1,
		      10:2,
		      11:2,
		      13:1, 17:1,
		      14:2,
		      15:2,
		      16:1 }
library['lu'] = { 0:2, 2:2,
		  1:2,
		  3:3, 6:3,
		  4:3,
		  5:3 }
library['gramschmidt'] = { 0:2, 2:2, 4:2,
			   1:1,
			   3:2,
			   5:2,
			   6:3, 9:3, 12:3,
			   7:3, 11:3,
			   8:3, 10:3, 13:3 }
library['dynprog'] = { 0:2,
		       1:3,
		       2:4,
		       3:4,
		       4:4,
		       5:4,
		       6:3,
		       7:3, 8:3 }
library['durbin'] = { 0:1, 1:1,
		      2:1, 3:1, 4:1, 10:1, 11:1, 16:1,
		      5:2, 15:2,
		      6:2,
		      7:2, 12:2,
		      8:2,
		      9:1, 17:1,
		      13:2,
		      14:2,
		      18:1,
		      19:1 }
library['adi'] = { 0:2, 2:2, 4:2, 5:2, 6:2, 8:2, 17:2, 19:2, 21:2, 22:2, 23:2, 25:2,
		   1:2, 3:2, 7:2,
		   9:2, 10:2, 11:2,
		   12:3,
		   13:3,
		   14:3,
		   15:3,
		   16:3,
		   18:2, 20:2, 24:2,
		   26:2, 27:2, 28:2,
		   29:3, 32:3, 33:3,
		   30:3,
		   31:3 }
library['fdtd-2d'] = { 0:2,
		       1:2,
		       2:3, 3:3, 5:3,
		       4:3,
		       6:3, 7:3, 9:3,
		       8:3,
		       10:3, 12:3, 14:3, 15:3,
		       11:3,
		       13:3 }
library['fdtd-apml'] = { 0:2, 5:2, 10:2, 17:2, 22:2, 46:2, 51:2, 56:2, 63:2, 68:2,
			 1:1, 4:1, 8:1, 11:1, 14:1, 20:1, 21:1, 54:1,
			 2:1,
			 3:1,
			 6:3, 7:3, 9:3, 53:3, 55:3,
			 12:2, 13:2, 16:2, 19:2, 58:2, 59:2, 62:2, 65:2,
			 15:2, 18:2, 61:2, 64:2,
			 23:1, 26:1, 28:1, 33:1, 40:1, 45:1, 69:1, 74:1, 79:1, 86:1, 91:1,
			 24:1, 27:1, 31:1, 34:1, 37:1, 43:1, 44:1,
			 25:1,
			 29:2, 30:2, 32:2,
			 35:1, 36:1, 39:1, 42:1, 52:1, 75:1, 76:1, 78:1, 81:1, 82:1, 85:1, 88:1,
			 38:2, 41:2, 84:2, 87:2,
			 47:3, 49:3, 50:3, 57:3, 60:3, 66:3, 67:3,
			 48:3,
			 70:2, 73:2, 77:2, 80:2, 83:2, 89:2, 90:2,
			 71:2, 72:2 }
library['jacobi-1d'] = { 0:2,
			       1:2, 3:2, 4:2, 5:2,
			       2:2 }
library['jacobi-2d'] = { 0:3, 5:3, 6:3, 7:3,
			       1:3,
			       2:3,
			       3:3,
			       4:3 }
library['seidel-2d'] = { 0:3,
			 1:3,
			 2:3,
			 3:3,
			 4:3,
			 5:3,
			 6:3,
			 7:3,
			 8:3,
			 9:3 }
library['correlation'] = { 0:1, 4:1, 5:1, 6:1, 11:1, 12:1, 13:1, 14:1, 15:1, 16:1,
			   1:2, 3:2, 7:2, 9:2, 10:2,
			   2:2, 8:2,
			   17:1, 19:1, 20:1, 22:1,
			   18:2, 21:2,
			   23:1,
			   24:2, 29:2,
			   25:3, 28:3,
			   26:3,
			   27:3,
			   30:2 }
library['covariance'] = { 0:1, 4:1, 5:1,
			  1:2, 2:2,
			  3:2,
			  6:1, 7:1,
			  8:2,
			  9:2, 15:2,
			  14:2,
			  10:3, 11:3,
			  12:3,
			  13:3 }
library['floyd-warshall'] = { 0:2, 3:2,
			      1:3,
			      2:3 }
library['reg_detect'] = { 0:3,
			  1:4,
			  2:3, 3:3,
			  4:4, 6:4,
			  5:4,
			  7:3,
			  8:3,
			  9:2, 10:2,
			  11:3, 13:3,
			  12:3 }
#library['exhaustive'] = { 0:3 }

def all_tests( size="standard", timing=False, debug=True ):
	"""This function is part of the system tests. It runs system_tests() for each
	test in the 'library' dictionary, and checks whether it returns True."""

	if timing: start_time = time.time()
	for test in library.keys():
		if not all_accesses( size=size, test=test, timing=timing, debug=debug ): return False
	if timing:
		stop_time = time.time()
		print "--- %s seconds ---" % (stop_time-start_time)

	return True

def all_accesses( size="standard", test="3mm", timing=False, debug=True ):
	"""This function is part of the system tests. It runs system_tests() for each
	access of a given test in the 'library' dictionary, and checks whether it
	returns True."""

	for access in library[test].keys():
		if not system_tests( size, test, access, timing=timing, debug=debug ): return False
	return True

def reconstruct_trace( d, c, cut ):
	"""This function is part of the system tests. It works by trying to reconstruct
	a trace specified as an array of first order differences with an incrementally
	deeper loop (variable "j"). This is done to ensure that traversal of the
	solution space is done in a breadth-first fashion, as described in the paper."""

	for j in xrange(1,cut+1):
		(sol_c, sol_i, sol_U, sol_w) = extract( d, c, cut=j )
		if not sol_c is None: return (sol_c, sol_i, sol_U, sol_w)
	return (None,None,None,None)

def system_tests( size="standard", test="3mm", access=0, timing=True, debug=False ):
	"""This function is part of the system tests. It generates a synthetic trace
	using generate_trace() and checks that it can be reconstructed using a loop
	of the exact level specified in the 'library' dictionary. It has two operation
	modifiers: when "timing" is enabled (by default), it runs the reconstruction
	measuring execution time using the "time" module. If "debug" is enabled it
	checks whether the reconstructed trace is exactly equal to the original one.
	Note that "debug" mode is very time consuming, as regenerating the trace from
	the obtained result is generally much more costly than the reconstruction
	process itself (which makes extensive use of NumPy vectorization features)."""

	print "Checking %s.%d..." % (test,access)
	print "Generating trace..."
	(a,c,d) = generate_trace( test=test, access=access, size=size )
	cut = library[test][access]
	if timing:
		start_time = time.time()

	print "Reconstructing..."
	(sol_c, sol_i, sol_U, sol_w) = reconstruct_trace( d,c, cut )

	if timing:
		stop_time = time.time()
		print "--- %s seconds ---" % (stop_time-start_time)

	if sol_c is None: 
		print "ERROR: solution not found"
		return False
	if not debug: return True
 
	print "Checking solution correctness..."

	# Only check that solution dimensionality agrees with the expected one
	# for "standard" size tests. Some smaller ones have lower-level representations.
	if len(sol_c) < cut and size == "standard":
		print "ERROR: %d-level solution found (expected %d-level)" % (len(sol_c), cut)
		return False

	ar=np.zeros_like(a)
	pos=0
	exec print_loop( 0, sol_c, sol_U, sol_w )
	if not np.array_equal( a, ar ):
		print "ERROR: original and generated traces differ"
		raise ValueError
		return False
	
	return True

def grow(d,c,I,start,cut,U,w):
	"""The grow() function, as described in the pseudo-code of Algorithm 2 in the
	paper."""

	len_c = len(c)
	if cut and len_c >= cut: return (None,None,None,None)

	# Add new stride for the new index to c vector
	# The new index can be added at any point in the stride array: recursively call
	# extract() for each insertion point, and ultimately return the smallest solution
	# (in terms of loop depth).
	for k in xrange(len_c+1): # Start inserting on the left, it's the most probable insertion point
        	# Check bounds and linearity for indices which loop when inserting this one
		new_U = np.zeros((len_c+1,len_c+1),dtype=np.int64)
        	new_U[:k,:k]=U[:k,:k]
	        new_U[:k,k+1:]=U[:k,k:]
	        new_U[k][k]=-1
	        new_U[k+1:,:k]=U[k:,:k]
	        new_U[k+1:,k+1:]=U[k:,k:]
	        new_w = np.zeros(len_c+1,dtype=np.int64)
	        new_w[:k]=w[:k]
	        new_w[k+1:]=w[k:]

	        new_I = np.zeros((len(I)+1,len(I[0])+1),dtype=np.int64)
	        new_I[:-1,:k] = I[:,:k]
	        new_I[:-1,k+1:] = I[:,k:]
	        new_I[-1,:k] = I[-1,:k]
	        new_I[-1,k] = 1
		Linear = True
		bounds_check_b( new_I, new_w, k )
		for x in xrange(k+1,len_c+1):
			if not bounds_check( new_I[:-1], new_U, new_w, x ):
				Linear = False
				break

		if not Linear: continue

	        new_c = np.zeros( len_c+1, dtype=np.int64 )
	        new_c[:k] = c[:k]
	        new_c[k] = d[start]+np.dot(c[k:],I[-1][k:]) # Stride of new index is the accumulated stride of looped indices plus observed first-order difference
	        new_c[k+1:] = c[k:]

		# Call extract() with the grown system. 
	        (new_c,new_I,new_U,new_w) = extract( d, new_c, new_I, start+1, cut, new_U, new_w )
	        if not new_c is None: return (new_c,new_I,new_U,new_w)

	# No solution found
	return (None,None,None,None)

def bounds_check_b( I, w, var_idx ):
	"""Check if the bounds vector needs to change as per Lemma 3.3."""

	if (I[-1,:var_idx] == 0).all(): w[var_idx] = I[-1][var_idx]

def bounds_check( I, U, w, l ):
	"""Check if the bounds matrix needs to change as described in Sec. 3.4.
	If so, build the (n-l+1) equation systems in Eq.(9) and solve it."""

	trip_count = np.dot( U[l], I[-1] ) + w[l]

	# If the expected trip count is not zero, then this index shouldn't be looping under the current bounds.
	if trip_count != 0:

		# Except if increasing l would yield a negative trip count for some inner index. In this case,
		# the current bounds are still legal.
		if trip_count > 0:
			alt_index = I[-1].copy()
			alt_index[l] += 1
			alt_index[l+1:] = 0 # alt_index holds the index derived from increasing l
			inner_trip_count = np.dot( U[l+1:], alt_index ) + w[l+1:]
			if (inner_trip_count < 0).any(): return True

		# If all coefficients in this row have already been calculated, then this solution is not valid.
		if not (U[l] == 0).any() : return False
		orig_row = U[l].copy()

		# Build the equation system in Eq.(9) in the paper. Note that since we start counting the rows at 0
		# in the code this system has l unknowns, instead of (l-1) as in the paper.

		# Extract l linearly independent indices from I. We apply an additional optimization not mentioned in the
		# paper due to lack of space. Since U is a lower triangular matrix only the first l elements matter. 
		# As such, we build matrix "i" (called tmp here to avoid confusion with I) as l x l, instead of n x l.
		#
		# This instruction extracts all indices in I such that its first l elements are not all zero.
		tmp = I[~(I[:,:l]==0).all(axis=1)]

		# If no such index exists, then it is not possible to build matrix i.
		if len(tmp) == 0: return False

		# Columns which are zero cannot be taken into account, as they cannot affect the calculation.
		# Remove them and fix the target rank accordingly.
		target_check = (tmp[:,:l]!=0).any(axis=0)
		target_indices = np.where( target_check )[0]
		tmp = tmp[:,target_indices]
		target_rank = len(target_indices)

		# Remove duplicates.
		idx = np.ones( len(tmp), dtype=np.bool )
		idx[1:] = (np.diff(tmp,axis=0) != 0).any(axis=1)
		tmp=tmp[idx]

		# If not enough indices to build matrix i, return.
		if len(tmp) < target_rank: return False

		# Now we have an array without duplicates. We need "target_rank" linearly independent rows, plus the
		# independent term to add to the RHS of the equations system.
		system = np.zeros( (target_rank,target_rank), dtype=np.int64 )
		coeffs = np.zeros( target_rank, dtype=np.int64 )
		system[0] = tmp[0]
		coeffs[0] = max(I[(I[:,target_indices]==system[0]).all(axis=1)][:,l]) - w[l] # Maximum for a specific combination of indices, as detailed in the paper. 

		# Add rows to the system until we run out of choices or we achieve "target_rank" linearly independent rows.
		rank = 1
		for k in xrange( 1, len(tmp) ):
			if rank == target_rank: break
			system[rank] = tmp[k]
			if np.linalg.matrix_rank(system) > rank:
				coeffs[rank] = max(I[(I[:,target_indices]==system[rank]).all(axis=1)][:,l]) - w[l]
				rank += 1

		# If we have achieved target rank, solve the system
		if rank == target_rank:
			if rank == 1: x = coeffs/system
			else: x = np.linalg.solve(system,coeffs)
			if x is None: return False # No linear solution
			else: U[l][target_indices] = np.transpose(x) # Solution found: copy to bounds matrix

		# If the calculated solution does not fix trip count, discard it
		if np.dot( U[l], I[-1] ) + w[l] != 0: return False

		# If any of the non-zero coefficients in the original row U[l] have changed, then this solution is not
		# valid, as it won't be sequential for some indices in I
		for j,x in enumerate(orig_row):
			if x != 0:
				if U[l][j] != x: return False

	return True

def mem_size( n, I, stack, U_stack, w_stack ):
	"""Calculate instantaneous memory consumption."""

	# Contributions to memory consumption:
	# - U is an inferior triangular matrix with main diagonal=-1, hence its size is n*(n-1)/2.
	# - Three vectors of size n: w, s and gamma.
	# - Vectors stored in I: len(I) * n
	# - Size of the index stack: 4 integers for each entry.
	# - Size of the U stack: A copy of U + 1 integer for each entry.
	# - Size of the w stack: A copy of w + 1 integer for each entry.

	U_size = n*(n-1)/2
	return U_size + n*3 + len(I)*n + len(stack) * 4 + len(U_stack)*(U_size+1) + len(w_stack) * (n+1)

with_gamma = 0
reached_j = 0
max_mem = 0
def extract(d,c,I=None,start=0,cut=0,U=None,w=None):
	"""The extract() function, as described in the pseudo-code of Algorithm 1 in the
	paper."""

	global reached_j, with_gamma, max_mem # Used for gamma counts (Table 2 in the paper) and memory size calculation

	# Initial U, w, and I as per Eq. (4). Note that c is supplied as input.
	if U is None: U=np.array([[-1]],dtype=np.int64)
	if w is None: w=np.array([1],dtype=np.int64)
	if I is None:
		I=np.array([[0],[1]],dtype=np.int64)
		reached_j = 0
		with_gamma = 0
		max_mem = 0

	# Backtracking stack. In order to save memory, it is divided into three different tables with cross-references: an index_stack,
	# which contains information about each backtracking point, and two different stacks for storing versions of U and w. In this way,
	# we avoid storing many duplicate copies of U and w into a single backtracking stack.
	index_stack = []
	U_stack = []
	w_stack = []
	save_index = -1

	len_c = len(c)	# Avoid calls to len()
	gamma = np.dot( U, I[-1] ) + w # Gamma: expected trip counts as described  in Sec. 3.3 (Eq.(6)).

	# Instantaneous memory consumption:
	max_mem = max( max_mem, mem_size( len_c, I, index_stack, U_stack, w_stack ) )
	j = start
	start_index = len_c
	while j < len(d):
		solved = False
		Linear = True

		# Try to determine branch priority as per Sec. 3.3: using gamma.
		gamma_idx = np.where(gamma>0)[0]
# AE: comment the len(gamma_idx) condition and uncomment the False one to deactivate gamma prediction (Table 1 in the paper)
# Note that, as detailed in the paper, under no gamma prediction most traces will NOT be reconstructed any time soon. We
# stopped each test after 48h.
		if len(gamma_idx):
#		if False:
			# Gamma suggests index k.
			k = gamma_idx[-1]
			accum = c[k] - np.dot(c[k+1:],I[-1,k+1:])

			# Check if stride produced by increasing index k is equal to the observed one.
			if accum == d[j]:
				solved = True
				if k == len_c-1:
					# Streamline: gamma predicts gamma[k] accesses. Try to process them at once.
					diffs = np.where( d[j:j+gamma[k]] != accum )[0]
					if len(diffs)>0 : inc = diffs[0]
					else: inc = gamma[k]
					if len(I) > save_index and len_c == cut: # If new indices can be potentially inserted to the right of this one, all indices need to be saved
							I[-1,-1] += inc
					else:
						# If observed increments don't agree with gamma prediction, we may need to recalculate bounds later.
						# Store the previous index as it may be needed to build the equation system in bounds_check().
						new_idx = I[-1].copy()
						new_idx[-1] += inc
						I = np.vstack( [I,new_idx] )
						max_mem = max( max_mem, mem_size( len_c, I, index_stack, U_stack, w_stack ) )

					j += inc

					# In most cases, gamma does not need to be recalculated using np.dot() (which is computationally costly). It is enough to decrement it
					# to account for the processed elements of the trace.
					gamma[-1] -= inc

					# Used for gamma prediction results.
					if reached_j <= j:
						with_gamma += inc
						reached_j = j
					else:
						with_gamma += max(0, j-reached_j)
						reached_j = max(reached_j, j)
				else:
					# Iterations of non-innermost indices are not processed in a streamlined fashion.
					if len(I) > save_index and (gamma[k+1:] == 0).all():
						# Increment index k by one unit, overwriting previous I[-1]
						I[-1,k] += 1
						I[-1,k+1:] = 0
					else:
						# Either this index needs to be saved for backtracking purposes or it may be needed later in bounds_check().
						# Add new element to I.
						new_idx = I[-1].copy()
						new_idx[k] += 1
						new_idx[k+1:] = 0
						I = np.vstack( [I,new_idx] )
						old_U = U.copy()
						for x in xrange(k+1,len_c):
							if not bounds_check( I[:-1], U, w, x ):
								Linear = False
								break
						if Linear:
							# If a new version of U has been generated which may be used by an element of the index_stack, store it into the U_stack.
							if index_stack and (not U_stack or index_stack[-1][0] > U_stack[-1][0]): 
								if not np.array_equal( U, old_U ): U_stack.append( (index_stack[-1][0], old_U) )
						else: U=old_U # If this solution is not linear, recover the old U.
						max_mem = max( max_mem, mem_size( len_c, I, index_stack, U_stack, w_stack ) )

					# In this case gamma[k] can just be decremented by one.
					# gamma[k+1:] need to be reset through np.dot().
					# gamma[:k] do not change.
					gamma[k] -= 1
					j += 1
					if reached_j <= j:
						with_gamma += 1
						reached_j += 1
					gamma[k+1:] = np.dot( U[k+1:], I[-1] ) + w[k+1:]

				if Linear: continue

		# Gamma prediction does not match observed stride. Try to see if there's and index that matches observed stride and fix bounds.
		if Linear:
			Bounds = True # Under very specific circumstances it is not necessary to recalculate bounds (just yet).
			accum = np.zeros( start_index, dtype=np.int64 )

			# Calculate the resulting strides of increasing each index
			for k in xrange( start_index-1, -1, -1): 
				accum[k] = c[k] - np.dot(c[k+1:],I[-1,k+1:])

			for k in xrange( start_index-1, -1, -1):
				# If one of the calculated strides equals the observed one, potential solution.
				if accum[k] == d[j]:
					one_accum = (accum==d[j]).sum() == 1 # Tests whether there's a single potential solution at this point.
					if (k == len_c-1) and (I[-1][:k] == 0).all():
						# Special case: discovery of w of innermost index. Streamline and insert special backtracking point.
						first_diff = np.argmax( d[j+1:] != accum[k] )
						if d[j+1:][first_diff] != accum[k]: new_j = first_diff + 1
						else: new_j = len(d)

						# index_limit is used to backtrack when streamlining. Instead of inserting one backtracking point
						# for each increment (first_diff points), we use index_limit.
						index_limit = I[-1][-1]+new_j-j
						index_stack.append( (len(I), j, k, index_limit, with_gamma) )
						save_index = len(I)
						new_I = I[-1].copy()
						new_I[k] = index_limit
						I = np.vstack( [I,new_I] )
						j = new_j
						reached_j = max( j, reached_j )
						max_mem = max( max_mem, mem_size( len_c, I, index_stack, U_stack, w_stack ) )
					else: # Increase selected index
						if (len_c == cut) and one_accum:
							# Under these circumstances it is not necessary to create a backtracking point, as there are no further solutions to explore
							# 1. len_c == cut => can't try to increase solution dimensionality
							# 2. one_accum => there are no other choices but to increase index k
							if (gamma[k+1:] == 0).all() and (len(I) > save_index) and (I[-1][:k] == 0).all() and (I[-1][k] > 1):
								# Under these circumstances it is not necessary to store the old index, as it will never be used by bounds_check():
								# 1. gamma[k+1:] == 0: no innermore index was expected to iterate => their dependences are explained by the current bounds
								# 2. len(I) > save_index: there is no backtracking point that needs this index for bounds_check()
								# 3. I[-1][:k] == 0: this index cannot be selected for bounds_check() in the future to determine dependences of index k
								# 4. I[-1][k] > 1: any possible dependence of innermore indices on k can be explained using previous entries in I
								I[-1,k] += 1
								I[-1,k+1:] = 0
								Bounds = False # Not necessary to check bounds at this point, either, as no new insights are gained by this increase
							else:
								new_I = I[-1].copy()
								new_I[k] += 1
								new_I[k+1:] = 0
								I = np.vstack( [I,new_I] )
								max_mem = max( max_mem, mem_size( len_c, I, index_stack, U_stack, w_stack ) )
						else:
							# A backtracking point needs to be created in case increasing k is not the correct choice.
							# In this case, it is always necessary to create a new I entry: in order to save memory, array I is not copied to the backtracking
							# stack. Instead, save_index is assigned the current length of I and it is guaranteed that entries below the current one are never
							# modified in the future.
							index_stack.append( (len(I), j, k, -1, with_gamma) )
							save_index = len(I)
							new_I = I[-1].copy()
							new_I[k] += 1
							new_I[k+1:] = 0
							I = np.vstack( [I,new_I] )
							max_mem = max( max_mem, mem_size( len_c, I, index_stack, U_stack, w_stack ) )
						j += 1
						reached_j = max(j, reached_j)
					solved = True
					break

		# If a solution was found, recalculate bounds (Sec. 3.4).
		if Linear and solved:
			start_index=len_c # start_index controls what solutions should be tried when backtracking.
			old_w = w.copy()
			bounds_check_b( I, w, k )
			if Bounds:
				old_U = U.copy()
				for x in xrange(len_c-1,k,-1):
					if not bounds_check( I[:-1], U, w, x ):
						Linear = False
						break
			if Linear:
				gamma[k:] = np.dot( U[k:], I[-1] ) + w[k:]
				if index_stack:
					# If a new version of U has been generated which may be used by an element of the index_stack, store it into the U_stack.
					if Bounds and not U_stack or index_stack[-1][0] > U_stack[-1][0]:
						if not np.array_equal(U,old_U): U_stack.append( (index_stack[-1][0], old_U) )
					# Same reasoning for w.
					if not w_stack or index_stack[-1][0] > w_stack[-1][0]:
						if not np.array_equal(w,old_w): w_stack.append( (index_stack[-1][0], old_w) )
				max_mem = max( max_mem, mem_size( len_c, I, index_stack, U_stack, w_stack ) )
				continue
			else:
				# If this solution is not linear, recover the old w and U.
				w = old_w
				U = old_U

		if Linear:
			# If no solution can be found through increasing a known index, try to add a new one
			(new_c,new_I,new_U,new_w) = grow( d, c, I, j, cut, U, w )
			if not new_c is None: return (new_c,new_I,new_U,new_w)

		if index_stack:
			# If no solution could be found yet, the only available option is to backtrack. Algorithm 1 in the paper does not use an explicit backtracking mechanism,
			# but rather uses recursive calls to extract(). That is not acceptable performance-wise.
			old_w = w.copy()
			old_U = U.copy()
			(len_I,j,start_index,index_limit,with_gamma) = index_stack.pop()

			if index_stack: save_index = index_stack[-1][0]
			else: save_index = -1

			I = I[:len_I] # recover matrix I

			# Recover matrix U, if necessary
			if U_stack and len(I) <= U_stack[-1][0]: U = U_stack.pop()[1]
			if w_stack and len(I) <= w_stack[-1][0]: w = w_stack.pop()[1]

			if index_limit != -1:
				# Special backtracking point for streamlined recognition
				index_limit -= 1
				j = index_limit + j - I[-1][-1]
				new_I = I[-1].copy()
				new_I[start_index] = index_limit
				I = np.vstack( [I,new_I] )
				bounds_check_b( I, w, start_index )

				# For performance, the standard version does limited backtracking. This is good enough for most traces (e.g., the entire PolyBench suite).
				# If an exhaustive backtracking is necessary to find a minimal solution, uncomment this line. An example of such a test is the synthetic "exhaustive"
				# test included in the generate_trace() function.
#				index_stack.append( (len(I),j,start_index,index_limit,with_gamma) )
			gamma = np.dot( U, I[-1] ) + w
			continue

		return ( None, None, None, None ) # No solution.

	for x in xrange(len_c):
		if not bounds_check( I, U, w, x ):
			return (None,None,None,None)

	print "Gamma/Total accesses: ", with_gamma, "/", len(d)+2
	print "Max. memory consumption: ", max_mem, " integers (", max_mem*4.0/1024," kB)"
	w[0] = I[-1][0]
	return (c,I,U,w)
